import pandas as pd
import numpy as np
from Bio import SeqIO
from Bio.Seq import Seq
from pandarallel import pandarallel
import matplotlib.pyplot as plt
from collections import Counter
import pickle
from datetime import date
from typeguard import typechecked
from os import listdir
from os.path import isfile, join
import re
from multiprocessing import Pool
from functools import partial
from itertools import repeat
import gzip
import random
from enum import Enum
from typing import Callable
from typing import Union, List, Mapping, Tuple, Optional, Any
from concurrent.futures import ProcessPoolExecutor

###
### UTILITY FUNCTIONS
###

def save_or_load_pickle(directory, label, py_object = None, date_string = None):
    '''Save a pickle for caching that is notated by the date'''
    
    if date_string == None:
        today = date.today()
        date_string = str(today.year) + ("0" + str(today.month) if today.month < 10 else str(today.month)) + str(today.day)
    
    filename = directory + label + "_" + date_string + '.pickle'
    print(filename)
    if py_object == None:
        with open(filename, 'rb') as handle:
            py_object = pickle.load(handle)
            return py_object
    else:
        with open(filename, 'wb') as handle:
            pickle.dump(py_object, handle, protocol=pickle.HIGHEST_PROTOCOL)

def display_all_pickle_versions(directory, label):
    '''Retrieve all pickles with a label, specifically to identify versions available'''
    return [f for f in listdir(directory) if isfile(join(directory, f)) and label == f[:len(label)]]


###
### GUIDE MAPPING HELPER FUNCTIONS
###

# Vector encoding of each base, since we want to vectorized
non_ambigious_encoding_dict = dict({
    "A": np.asarray([1,0,0,0]),
    "C": np.asarray([0,1,0,0]),
    "G": np.asarray([0,0,1,0]),
    "T": np.asarray([0,0,0,1]),
    "U": np.asarray([0,0,0,1])
})

'''
 Helper function for retrieving the encoding for ambigious bases based on IUPAC codes
'''
def ambiguity_encoder(bases):
    return np.logical_or.reduce([non_ambigious_encoding_dict[base] for base in bases]).astype(int)

'''
    Final dictionary for getting encoding of each IUPAC base
'''
full_encoding_dict = dict({
    "A": non_ambigious_encoding_dict["A"],
    "C": non_ambigious_encoding_dict["C"],
    "G": non_ambigious_encoding_dict["G"],
    "T": non_ambigious_encoding_dict["T"], 
    "R": ambiguity_encoder(["A", "G"]),
    "Y": ambiguity_encoder(["C", "T"]),
    "S": ambiguity_encoder(["G", "C"]),
    "W": ambiguity_encoder(["A", "T"]),
    "K": ambiguity_encoder(["G", "T"]),
    "M": ambiguity_encoder(["A", "C"]),
    "B": ambiguity_encoder(["C", "G", "T"]),
    "D": ambiguity_encoder(["A", "G", "T"]),
    "H": ambiguity_encoder(["A", "C", "T"]),
    "V": ambiguity_encoder(["A", "C", "G"]),
    "N": ambiguity_encoder(["A", "C", "G", "T"]),
})

'''
    Main function to encode a single base
'''
def encode_DNA_base(char):
    return full_encoding_dict[char]
encode_DNA_base_vectorized = np.vectorize(encode_DNA_base, signature='()->(n)') # Vectorized function for a string (i.e. gRNA)

'''
    Function for converting string (i.e. gRNA) into a np array of chars  - may be deprecated (NOTE 20221202)
'''
def numpify_string(string):
    return np.array(list(string), dtype=str)
numpify_string_vectorized = np.vectorize(numpify_string, signature='()->(n)') # Vectorize the function

def encode_guide_series(guide_series) -> np.array:
    guide_numpy = guide_series.to_numpy(dtype=object)
    guide_numpy = guide_numpy.astype(str)
    guide_numpy_char = np.array(list(map(list, guide_numpy))) # Map into a list of list of characters
    guide_numpy_encoding = encode_DNA_base_vectorized(guide_numpy_char)
    return guide_numpy_encoding


def retrieve_hamming_distance_whitelist(target_guide_encoded, whitelist_guide_encoded):
    '''
        This takes a encoded guide sequence and a list of encoded whitelisted guides and matrix computes the hamming distance of the 
        encoded guide across all whitelisted guides in a single operation
        
        (target_guide_encoded*whitelist_guide_encoded[:, np.newaxis]).sum(axis=3) # Determines 
        
    '''
    return ((target_guide_encoded*whitelist_guide_encoded[:, np.newaxis]).sum(axis=3)^1).sum(axis=2).flatten()

def determine_hamming_distance_classic(seq1, seq2):
    if len(seq1) != len(seq2):
        raise ValueError(f"Sequences must have equal length. {seq1} and {seq2}")
    
    distance = 0
    for i in range(len(seq1)):
        if seq1[i] != seq2[i]:
            distance += 1
    
    return distance

###
### GUIDE PARSING HELPER FUNCTIONS
###

'''
    Strategies to parse read
'''
@typechecked
def parse_read_positional(read_sequence: Union[str, Seq], position_start: int, position_end: int) -> Union[str, Seq]:  
    return read_sequence[position_start:position_end]

@typechecked
def parse_read_left_flank(read_sequence: Union[str, Seq], left_flank:Union[str, Seq], guide_sequence_length:int) -> Union[str, Seq]: 
    position_start = read_sequence.find(left_flank) + len(left_flank)
    return read_sequence[position_start:position_start+guide_sequence_length]

@typechecked
def parse_read_right_flank(read_sequence: Union[str, Seq], right_flank:Union[str, Seq], guide_sequence_length:int) -> Union[str, Seq]:
    position_end = read_sequence.find(right_flank)
    return read_sequence[position_end-guide_sequence_length:position_end]


'''
    Extract the guide sequence from the read provided
'''
@typechecked
def parse_guide_sequence(read_sequence: Union[str, Seq], parser_function: Callable) -> Union[str, Seq]:
    read_guide_sequence = parser_function(read_sequence)
    return read_guide_sequence


'''
    Iterate over all the reads in the FASTQ (parallelized) and retrieve the observed guide sequence
'''
@typechecked
def retrieve_fastq_guide_sequences(fastq_file: str, parse_left_flank: bool = True, parse_flank_sequence: Union[None, str] = None, cores: int=1) -> Union[List[str], List[Seq]]:
    parse_guide_sequence_p = None
    if parse_left_flank:
        if parse_flank_sequence is None:
            print("No flank sequence passed. Setting left-flank default sequence to CACCG assuming U6 G+N20 guide")
            flank_sequence = "CACCG"
        else:
            flank_sequence = parse_flank_sequence

        parse_read_left_flank_p = partial(parse_read_left_flank, left_flank=flank_sequence, guide_sequence_length=20)
        parse_guide_sequence_p = partial(parse_guide_sequence, parser_function=parse_read_left_flank_p)
    else:
        if parse_flank_sequence is None:
            print("No flank sequence passed. Setting right-flank default sequence to GTTTT. If you are using sgRNA(F+E) design, flank sequence may need to be changed")
            flank_sequence = "GTTTT"
        else:
            flank_sequence = parse_flank_sequence
        
        parse_read_right_flank_p = partial(parse_read_right_flank, right_flank=flank_sequence, guide_sequence_length=20)
        parse_guide_sequence_p = partial(parse_guide_sequence, parser_function=parse_read_right_flank_p)

    def parse_fastq_entry(seq):
        return parse_guide_sequence_p(seq.seq)

    
    read_executor_cores = int(np.sqrt(cores))
    parse_executor_cores = int(np.sqrt(cores))
    read_executor_chunk_size = 50000
    parse_executor_chunk_size = int(read_executor_chunk_size/parse_executor_cores)

    def process_fastq_chunk(chunk):
        with ProcessPoolExecutor(max_workers=parse_executor_cores) as executor:
            return list(executor.map(parse_fastq_entry, chunk, chunksize=parse_executor_chunk_size))
    
    with ProcessPoolExecutor(max_workers=read_executor_cores) as executor:
        if fastq_file.endswith('.gz'):
            print(f"Opening FASTQ.gz file with gzip, filename={fastq_file}")
            with gzip.open(fastq_file, "rt", encoding="utf-8") as handle:
                print("FASTQ opened")
                fastq_guide_sequences = list(executor.map(process_fastq_chunk, SeqIO.parse(handle, 'fastq'), chunksize = read_executor_chunk_size))
        else:
            fastq_guide_sequences = list(executor.map(process_fastq_chunk, SeqIO.parse(fastq_file, 'fastq')))

    return fastq_guide_sequences

###
### GUIDE MAPPING MAIN FUNCTIONS
###
class GuideCountError(Enum):
    NO_MATCH = "No guide found within hamming distance"
    MULTIPLE_MATCH = "Multiple exact matches found for guide (likely a truncated guide read assuming guide series is unique)"
    NO_GUIDE_WITH_SAME_LENGTH = "No whitelisted guides with same length as observed guide - maybe try enabling truncating whitelisted guides"
    NO_BARCODE_SAME_LENGTH = "No whitelisted barcode with the same length as the observed"
    NO_MATCH_GUIDE_HAMMING_THRESHOLD = "No whitelisted guide that is below the provided guide hamming distance threshold"
    NO_MATCH_BARCODE_HAMMING_THRESHOLD = "No whitelisted barcode that is below the provided barcode hamming distance threshold"
    NO_MATCH_SURROGATE_HAMMING_THRESHOLD = "The inferred whitelisted surrogate does not match with the observed surrogate below the given hamming distance threshold"
    MULTIPLE_MATCH_EXACT = "Multiple exact guide matches, double check that there are no duplicates in your guide library (especially if truncation is enabled)"
    NO_MATCH_MISSING_INFO = "The guide/surrogate/barcode were not all provided"
    NO_MATCH_OBSERVED_SURROGATE_SHORTER = "The observed surrogate is shorter than the inferred surrogate"
    
   
'''
    Given an observed, potentially self-edited guide (in 'row["observed_sequence"]'), try and find the true guide sequence from the whitelist ("guide_sequences_series") based on hamming distance
    # TODO (3/1/2023): This is the main bottleneck in terms of computational efficiency, see if it can be sped up
'''

@typechecked
def infer_true_guides(observed_guide_sequence: str, whitelist_guide_sequences_series: Union[List[str], pd.Series],
encoded_whitelist_guide_sequences_series, consider_truncated_sequences: bool = False, hamming_threshold: int = 3, verbose_result = False):
    #observed_guide_sequence = str(observed_guide_row["observed_sequence"])
    
    # If considering truncated guides, truncate all whitelist guides to the same size of the observed guide, else only consider whitelisted guides of the same length (likely all guides provided are same length of 20nt)
    if consider_truncated_sequences == True:
        whitelist_guide_sequences_series = whitelist_guide_sequences_series.apply(lambda guide: guide[0:len(observed_guide_sequence)])
    else:
        whitelist_guide_sequences_series = whitelist_guide_sequences_series[whitelist_guide_sequences_series.str.len() == len(observed_guide_sequence)]
        if len(whitelist_guide_sequences_series) == 0:
            return GuideCountError.NO_GUIDE_WITH_SAME_LENGTH 
        

    # Determine if there are exact matches, hopefully just a single match
    whitelist_guide_sequences_series_match = whitelist_guide_sequences_series[whitelist_guide_sequences_series == observed_guide_sequence]
    
    # If there is a single exact match, great, no need for fancy mat
    if len(whitelist_guide_sequences_series_match) == 1: # Exact match found, return
        return tuple([str(whitelist_guide_sequences_series_match.index[0])])
    
    # If no matches, possible a self-edit or a sequencing error, try and find guide with lowest hamming distance
    elif len(whitelist_guide_sequences_series_match) == 0: # No match found, search based on hamming distance
        
        # Encode the whitelisted guides 
        # NOTE 20221202: Potentially improve efficiency by passing in the encoded guide series (assuming no truncation) so that this does not have to be ran on every guide
        #guide_sequences_series_encoded = encode_guide_series(whitelist_guide_sequences_series_match)
        
        # Encode the observed guide
        try:
            observed_guide_sequence_encoded = encode_DNA_base_vectorized(numpify_string_vectorized(observed_guide_sequence)) 
        except Exception as e:
            print(f"Error on observed guide sequence: {observed_guide_sequence}")
            raise e
        # Calculate the hamming distance of the guide with all whitelisted guides - vectorized operation
        observed_guide_sequence_dists = retrieve_hamming_distance_whitelist(observed_guide_sequence_encoded, encoded_whitelist_guide_sequences_series)
        
        # Get the minimum hamming distance calculated
        hamming_min = observed_guide_sequence_dists.min()
        
        # Get all whitelisted guides with the minimum hamming distance (could be multiple)
        guides_with_hamming_min = whitelist_guide_sequences_series[np.where(observed_guide_sequence_dists == hamming_min)[0]]
        
        # If the minimum hamming distance is greater than the specified threshold, then the guide is too ambigious to assign, so no match.
        if hamming_min >= hamming_threshold:
            if verbose_result:
                return {"Error": GuideCountError.NO_MATCH_GUIDE_HAMMING_THRESHOLD, "hamming_min": hamming_min}
            else:
                return GuideCountError.NO_MATCH_GUIDE_HAMMING_THRESHOLD
        
        # If there are multiple guides with the minimum hamming distance, then the guide is ambigious, so no mapping (due to multiple match)
        # NOTE (3/30/2023): This could potentially be an issue with igRNAs maybe?
        elif len(guides_with_hamming_min) > 1:
            if verbose_result:
                return {"Error": GuideCountError.MULTIPLE_MATCH,
                "exact_match": False, 
                "num_matches": len(guides_with_hamming_min),
                "matches": guides_with_hamming_min,
                "hamming_min": hamming_min}
            else:
                return GuideCountError.MULTIPLE_MATCH
        
        # Else if there is 1 guide with the match, then return the match
        else:
            return tuple([str(guides_with_hamming_min.index[0])])
    
    # Else if there are multiple exact match, which should never occur unless the whitelisted guide list is not unique, then return multiple match.
    else:
        
        if verbose_result:
            return {"Error": GuideCountError.MULTIPLE_MATCH_EXACT,
            "exact_match": True, 
            "num_matches": len(whitelist_guide_sequences_series_match),
            "matches": whitelist_guide_sequences_series_match,
            "hamming_min": 0}
        else:
            return GuideCountError.MULTIPLE_MATCH_EXACT
        #raise Exception("Multiple exact matches of the provided whitelisted guides - there are likely duplicates in the provided whitelist, please remove. Observed guide={}, guide matches={}".format(observed_guide_sequence, guide_sequences_series_match)) # NOTE 12/6/22: REMOVED THIS EXCEPTION - another reason is from truncated guides having multiple matches. In production code, just make sure to ensure that the whitelist is the set.

@typechecked
def infer_true_reporters(observed_reporter_sequences, whitelist_guide_reporter_df: pd.DataFrame,
encoded_whitelist_guide_sequences_series, encoded_whitelist_barcodes_series, surrogate_hamming_threshold: int = 10, barcode_hamming_threshold: int = 2, hamming_threshold: int = 7, verbose_result = False):
    for element in observed_reporter_sequences:
        if (element is None) or (element == "None") or (element == ""):
            if verbose_result:
                return {"Error": GuideCountError.NO_MATCH_MISSING_INFO, "message": "Observed protospacer/surrogate/barcode is None"}
            else:
                return GuideCountError.NO_MATCH_MISSING_INFO
    
    observed_reporter_sequences = pd.Series(observed_reporter_sequences, index=["protospacer", "surrogate", "barcode"])
    
    # Check if the observed protospacer length is the same length of the whitelisted guides
    if len(observed_reporter_sequences["protospacer"]) != encoded_whitelist_guide_sequences_series.shape[1]: 
        if verbose_result:
            return {"Error": GuideCountError.NO_GUIDE_WITH_SAME_LENGTH, "message": f"Observed protospacer {observed_reporter_sequences['protospacer']} not of correct length: {encoded_whitelist_guide_sequences_series.shape[1]}"}
        else:
            return GuideCountError.NO_GUIDE_WITH_SAME_LENGTH
    if len(observed_reporter_sequences["barcode"]) != encoded_whitelist_barcodes_series.shape[1]: 
        if verbose_result:
            return {"Error": GuideCountError.NO_BARCODE_SAME_LENGTH, "message": f"Observed barcode {observed_reporter_sequences['barcode']} not of correct length: {encoded_whitelist_barcodes_series.shape[1]}"}
        else:
            return GuideCountError.NO_BARCODE_SAME_LENGTH   
            
    # Determine if there are exact matches, hopefully just a single match
    whitelist_guide_reporter_df_match = whitelist_guide_reporter_df[(whitelist_guide_reporter_df["protospacer"]==observed_reporter_sequences["protospacer"]) & (whitelist_guide_reporter_df["surrogate"]==observed_reporter_sequences["surrogate"]) & (whitelist_guide_reporter_df["barcode"]==observed_reporter_sequences["barcode"])] #whitelist_guide_sequences_series == observed_guide_sequence
    
    # If there is a single exact match, great, no need for fancy math
    if whitelist_guide_reporter_df_match.shape[0] == 1: # Exact match found, return
        return tuple(whitelist_guide_reporter_df_match.iloc[0])
    
    # If no matches, possible a self-edit or a sequencing error, try and find guide with lowest hamming distance
    elif whitelist_guide_reporter_df_match.shape[0] == 0: # No match found, search based on hamming distance
        
        ###
        ### FILTER BY BARCODE (NOTE: 4/2/23: This is the main difference with the traditional filtering by protospacer)
        ### TODO 4/2/23: Assumes perfect barcode match, but in the future I can also select for barcodes of 1 hamming - 2 hamming if there is no matches with 0 hamming
        ###
        observed_barcode_encoded = encode_DNA_base_vectorized(numpify_string_vectorized(observed_reporter_sequences["barcode"]))  # Encode the observed barcode
        observed_barcode_dists = retrieve_hamming_distance_whitelist(observed_barcode_encoded, encoded_whitelist_barcodes_series) # Retrieve hamming distance with whitelist barcode
        barcode_hamming_min = observed_barcode_dists.min() # Get the barcode with the minimum  hamming distance
        if barcode_hamming_min >= barcode_hamming_threshold: # If the barcode surpasses the threshold, fail the read due to no barcode. 
            if verbose_result:
                return {"Error": GuideCountError.NO_MATCH_BARCODE_HAMMING_THRESHOLD, "barcode_hamming_min": barcode_hamming_min, "message": f"No barcode below threshold {barcode_hamming_threshold}"}
            else:
                return GuideCountError.NO_MATCH_BARCODE_HAMMING_THRESHOLD
            
        barcode_matches_indices = np.where(observed_barcode_dists == barcode_hamming_min)[0] # Get the indices of ALL barcode matches
        
        whitelist_guide_reporter_df_barcode = whitelist_guide_reporter_df.iloc[barcode_matches_indices] # Get the whitelist reporters with the matched barcode(s)
        encoded_whitelist_guide_sequences_series_barcode = encoded_whitelist_guide_sequences_series[barcode_matches_indices]
        
        ###
        ### Map the protospacer among those with the correct barcode
        ###
        observed_guide_sequence_encoded = encode_DNA_base_vectorized(numpify_string_vectorized(observed_reporter_sequences["protospacer"]))  # Encode the observed protospacer
        
        
        observed_guide_sequence_dists = retrieve_hamming_distance_whitelist(observed_guide_sequence_encoded, encoded_whitelist_guide_sequences_series_barcode) # Calculate the hamming distance of the guide with all whitelisted guides - vectorized operation
        
        
        hamming_min = observed_guide_sequence_dists.min() # Get the minimum hamming distance calculated
        
        reporters_with_hamming_min_df = whitelist_guide_reporter_df_barcode.iloc[np.where(observed_guide_sequence_dists == hamming_min)[0]] # Get all whitelisted guides with the minimum hamming distance (could be multiple)
        
        if hamming_min >= hamming_threshold: # If the minimum hamming distance is greater than the specified threshold, then the guide is too ambigious to assign, so no match.
            if verbose_result:
                return {"Error": GuideCountError.NO_MATCH_GUIDE_HAMMING_THRESHOLD, "hamming_min": hamming_min}
            else:
                return GuideCountError.NO_MATCH_GUIDE_HAMMING_THRESHOLD
        
        # If there are multiple guides with the minimum hamming distance, then the guide is ambigious, so no mapping (due to multiple match)
        # NOTE (3/30/2023): This could potentially be an issue with igRNAs maybe? (Response 6/30/23) No, the barcode should be able to distinguish between igRNAs
        elif reporters_with_hamming_min_df.shape[0] > 1:
            if verbose_result:
                return {"Error": GuideCountError.MULTIPLE_MATCH,
                "exact_match": False, 
                "num_matches": reporters_with_hamming_min_df.shape[0],
                "matches": reporters_with_hamming_min_df,
                "hamming_min": hamming_min}
            else:
                return GuideCountError.MULTIPLE_MATCH
        
        # Else if there is 1 guide with the match, then double check that the observed surrogate matches the mapped surrogate (or if it is due to recombination)
        else:
            inferred_reporter_sequences = tuple(reporters_with_hamming_min_df.iloc[0])
            inferred_surrogate_sequence = inferred_reporter_sequences[1]
            observed_surrogate_sequence = observed_reporter_sequences["surrogate"]
            
            if len(observed_surrogate_sequence) >= len(inferred_surrogate_sequence):
                observed_surrogate_sequence = observed_surrogate_sequence[-len(inferred_surrogate_sequence):] # Because their may be slippage of the polyT upstream of surrogate, slice relative to the downstream end.
                surrogate_hamming_distance = determine_hamming_distance_classic(inferred_surrogate_sequence, observed_surrogate_sequence)
                if surrogate_hamming_distance >= surrogate_hamming_threshold:
                    if verbose_result:
                        return {"Error": GuideCountError.NO_MATCH_SURROGATE_HAMMING_THRESHOLD, "surrogate_hamming_distance": surrogate_hamming_distance, "inferred_reporter_sequences": inferred_reporter_sequences}
                    else:
                        return GuideCountError.NO_MATCH_SURROGATE_HAMMING_THRESHOLD
                else:
                    return inferred_reporter_sequences
            else:
                if verbose_result:
                        return {"Error": GuideCountError.NO_MATCH_OBSERVED_SURROGATE_SHORTER, "inferred_reporter_sequences": inferred_reporter_sequences}
                else:
                    return GuideCountError.NO_MATCH_OBSERVED_SURROGATE_SHORTER
    
    # Else if there are multiple exact match, which should never occur unless the whitelisted guide list is not unique, then return multiple match.
    else:
        if verbose_result:
            return {"Error": GuideCountError.MULTIPLE_MATCH_EXACT,
            "exact_match": True, 
            "num_matches": whitelist_guide_reporter_df_match.shape[0],
            "matches": whitelist_guide_reporter_df_match,
            "hamming_min": 0}
        else:
            return GuideCountError.MULTIPLE_MATCH_EXACT
        #raise Exception("Multiple exact matches of the provided whitelisted guides - there are likely duplicates in the provided whitelist, please remove. Observed guide={}, guide matches={}".format(observed_guide_sequence, guide_sequences_series_match)) # NOTE 12/6/22: REMOVED THIS EXCEPTION - another reason is from truncated guides having multiple matches. In production code, just make sure to ensure that the whitelist is the set.

     
'''
    This performs simulation of determining how many mutations it takes for a guide to be ambigiously mapped based on hamming distance.

    This is performed on *sample_count* number of randomly selected guides, and the distribution of mutation count is determined. The 5% quantile is used as the threshold.
    
    This is useful in determing the ideal hamming distance threshold specific to a guide library
'''
@typechecked
def determine_hamming_threshold(whitelist_guide_sequences_series: Union[List[str],pd.Series], encoded_whitelist_guide_sequences_series, sample_count: int = 100, quantile: float = 0.05) -> float:
    #encoded_whitelist_guide_sequences_series = encode_guide_series(guide_sequences_series)
    
    mutation_count_until_nonunique = []
    for i in range(sample_count):
        # Sample a guide from whitelist
        sampled_guide = whitelist_guide_sequences_series.sample()[0]
        
        # Generate order of base positions to mutate sequentially.
        guide_position_list = list(range(len(sampled_guide)))
        random.shuffle(guide_position_list)
        
        # Create temporary variable to represent mutated guide
        current_guide_sequence = sampled_guide
        
        # Iterate through positions to mutate
        for iteration, position in enumerate(guide_position_list):
            
            # Mutate the guide
            current_guide_sequence_separated = list(current_guide_sequence)
            guide_position_nt = current_guide_sequence_separated[position]
            nt_list = ["A", "C", "T", "G"]
            nt_list.remove(guide_position_nt.upper())
            new_nt = random.sample(nt_list, 1)[0]
            current_guide_sequence_separated[position] = new_nt
            current_guide_sequence = "".join(current_guide_sequence_separated)
            current_guide_sequence_encoded = encode_DNA_base_vectorized(numpify_string_vectorized(current_guide_sequence)) 
            
            # Calculate the hamming distance of the mutated guide to all other guides
            hamming_distances =  retrieve_hamming_distance_whitelist(current_guide_sequence_encoded, encoded_whitelist_guide_sequences_series)
            if len(np.where(hamming_distances == hamming_distances.min())[0]) > 1: # TODO (3/30/23): Can also add to the conditional whether the minimum hamming distance guide is still the original guide - but it is probably rare cases where this would not be the case, so not too important to implement
                mutation_count_until_nonunique.append(iteration+1)
                break   
    mutation_count_until_nonunique = pd.Series(mutation_count_until_nonunique)
    
    # From the mutation count distribution, calculate the threshold based on the provided quantile.
    hamming_distance_threshold = mutation_count_until_nonunique.quantile(quantile)
    return hamming_distance_threshold


'''
    Take in input FASTQ filename, and a set of whitelisted guide sequences
'''
@typechecked
def get_guide_counts(observed_guide_sequences_counts: Counter, whitelist_guide_sequences_series: pd.Series, hamming_threshold_strict: int = 3, hamming_threshold_dynamic: bool = False, cores: int=1):

    whitelist_guide_sequences_series.index = whitelist_guide_sequences_series.values
    # Create observed guide DF to contain all information
    # The "observed sequence" column represents all *unique* observed sequences that may have self-edits/errors. The "observed counts" column represents the count of each observed count.
    observed_guides_df = pd.DataFrame({"observed_sequence":[str(sequence) for sequence in observed_guide_sequences_counts.keys()], "observed_counts":observed_guide_sequences_counts.values()})
 
    # Get the hamming distance threshold. THe hamming distance must be below this threshold to assign an observed guide to a whitelist guide.
    encoded_whitelist_guide_sequences_series = encode_guide_series(whitelist_guide_sequences_series)
    if hamming_threshold_dynamic:
        hamming_threshold = int(determine_hamming_threshold(whitelist_guide_sequences_series, encoded_whitelist_guide_sequences_series, sample_count = 100, quantile = 0.05))
        print("Hamming threshold is " + str(hamming_threshold))
    else:
        hamming_threshold = hamming_threshold_strict
        
    # Infer whitelist guides from observed guides
    print("Inferring the true guides from observed guides")

    infer_true_guides_p = partial(infer_true_guides,
            whitelist_guide_sequences_series=whitelist_guide_sequences_series,
            encoded_whitelist_guide_sequences_series=encoded_whitelist_guide_sequences_series,
            hamming_threshold=hamming_threshold, verbose_result=False)

    inferred_true_guide_sequences = None
    
    if cores > 1:
        with Pool(cores) as pool:
            print("Inferencing parallel")
            inferred_true_guide_sequences = pool.map(
            infer_true_guides_p,
            observed_guides_df["observed_sequence"]
            )
    else:
        print("Inferencing sequentially")
        inferred_true_guide_sequences = [infer_true_guides_p(observed_sequence) for observed_sequence in observed_guides_df["observed_sequence"]]    
    
    print("Completed inference")

    observed_guides_df["inferred_guides"] = inferred_true_guide_sequences
    
    '''
        QC
    '''
    # QC: Determine the number of guides passed by determining that the result is not an error
    observed_guides_df_passed_inference = observed_guides_df[observed_guides_df["inferred_guides"].apply(lambda guide : type(guide) != GuideCountError)]
    # QC: Calculate number of guides that were unassigned
    observed_guides_df_no_match = observed_guides_df[observed_guides_df["inferred_guides"].apply(lambda guide : guide == GuideCountError.NO_MATCH)]
    # QC: Calculate number of guides with multiple inferred guides
    observed_guides_df_multiple_match = observed_guides_df[observed_guides_df["inferred_guides"].apply(lambda guide : guide == GuideCountError.MULTIPLE_MATCH)]
    # QC: Calculate percent mapped
    percent_mapped = observed_guides_df_passed_inference["observed_counts"].sum()/observed_guides_df["observed_counts"].sum()
    
    # Retrieve the observed sequences that were mapped and set the inferred guides
    inferred_guide_sequence_counter = Counter()
    for _, row in observed_guides_df_passed_inference.iterrows():
        inferred_guide_sequence_counter[row["inferred_guides"]] += row["observed_counts"]
    
    whitelist_guide_sequences_series_counts = whitelist_guide_sequences_series.apply(lambda guide: inferred_guide_sequence_counter[guide])
    whitelist_guide_sequences_series_counts.index = whitelist_guide_sequences_series
    
    qc_dict = {"guide_sequences_unassigned_counts":observed_guides_df_no_match["observed_counts"].sum(), "guide_sequences_multiple_counts": observed_guides_df_multiple_match["observed_counts"].sum(), "total_guide_counts": observed_guides_df["observed_counts"].sum(), "percent_mapped": percent_mapped}
    
    return whitelist_guide_sequences_series_counts, observed_guides_df, qc_dict

@typechecked
def get_reporter_counts(observed_guide_reporters_counts: Counter, whitelist_guide_reporter_df: pd.DataFrame, surrogate_hamming_threshold_strict: int = 2, barcode_hamming_threshold_strict: int = 2, hamming_threshold_strict: int = 7, hamming_threshold_dynamic: bool = False, cores: int=1):

    whitelist_guide_reporter_df.index = whitelist_guide_reporter_df["protospacer"]
    
    # Create observed guide DF to contain all information
    # The "observed sequence" column represents all *unique* observed sequences that may have self-edits/errors. The "observed counts" column represents the count of each observed count.
    observed_guides_df = pd.DataFrame({"observed_sequence":observed_guide_reporters_counts.keys(), "observed_counts":observed_guide_reporters_counts.values()})
 
    # Get the hamming distance threshold. THe hamming distance must be below this threshold to assign an observed guide to a whitelist guide.
    encoded_whitelist_guide_sequences_series = encode_guide_series(whitelist_guide_reporter_df["protospacer"])
    encoded_whitelist_barcodes_series = encode_guide_series(whitelist_guide_reporter_df["barcode"])
    
    if hamming_threshold_dynamic:
        hamming_threshold = int(determine_hamming_threshold(whitelist_guide_reporter_df["protospacer"], encoded_whitelist_guide_sequences_series, sample_count = 100, quantile = 0.05))
        print("Hamming threshold is " + str(hamming_threshold))
    else:
        hamming_threshold = hamming_threshold_strict
        
    # Infer whitelist guides from observed guides
    print("Inferring the true guides from observed guides")

    infer_true_reporters_p = partial(infer_true_reporters,
            whitelist_guide_reporter_df=whitelist_guide_reporter_df,
            encoded_whitelist_guide_sequences_series=encoded_whitelist_guide_sequences_series,
            encoded_whitelist_barcodes_series=encoded_whitelist_barcodes_series,
            hamming_threshold=hamming_threshold, barcode_hamming_threshold=barcode_hamming_threshold_strict, surrogate_hamming_threshold=surrogate_hamming_threshold_strict, verbose_result=False)

    inferred_true_reporter_sequences = None
    if cores > 1:
        with Pool(cores) as pool:
            inferred_true_reporter_sequences = pool.map(
            infer_true_reporters_p,
            observed_guides_df["observed_sequence"]
           )
    else:
        inferred_true_reporter_sequences = [infer_true_reporters_p(obs_reporter) for obs_reporter in observed_guides_df["observed_sequence"]]
    
    print("Completed inference")

    observed_guides_df["inferred_guides"] = inferred_true_reporter_sequences
    
    '''
        QC
    '''
    print("Retrieving QC tables")
    # QC: Determine the number of guides passed by determining that the result is not an error
    observed_guides_df_passed_inference = observed_guides_df[observed_guides_df["inferred_guides"].apply(lambda guide : type(guide) != GuideCountError)]
    # QC: Calculate number of guides that were unassigned
    observed_guides_df_no_match = observed_guides_df[observed_guides_df["inferred_guides"].apply(lambda guide : guide == GuideCountError.NO_MATCH)]
    # QC: Calculate number of guides with multiple inferred guides
    observed_guides_df_multiple_match = observed_guides_df[observed_guides_df["inferred_guides"].apply(lambda guide : guide == GuideCountError.MULTIPLE_MATCH)]
    # QC: Calculate percent mapped
    percent_mapped = observed_guides_df_passed_inference["observed_counts"].sum()/observed_guides_df["observed_counts"].sum()
    
    print("Get whitelist reporter counts")
    # Retrieve the observed sequences that were mapped and set the inferred guides
    inferred_guide_sequence_counter = observed_guides_df_passed_inference.groupby("inferred_guides")["observed_counts"].sum()

    whitelist_guide_reporter_counts = whitelist_guide_reporter_df.apply(lambda reporter: inferred_guide_sequence_counter.get(tuple(reporter), 0), axis=1)
    
    
    multi_index = pd.MultiIndex.from_arrays([whitelist_guide_reporter_df['protospacer'], whitelist_guide_reporter_df['surrogate'], whitelist_guide_reporter_df['barcode']], names=['protospacer', 'surrogate', 'barcode'])
    whitelist_guide_reporter_counts.index = multi_index
    

    qc_dict = {"guide_sequences_unassigned_counts":observed_guides_df_no_match["observed_counts"].sum(), "guide_sequences_multiple_counts": observed_guides_df_multiple_match["observed_counts"].sum(), "total_guide_counts": observed_guides_df["observed_counts"].sum(), "percent_mapped": percent_mapped}
    
    return whitelist_guide_reporter_counts, observed_guides_df, qc_dict



###
### Read in sequences from a pre-parsed reporter file
### 
import csv
import multiprocessing as mp
from collections import Counter
from functools import reduce

def gen_chunks(reader, chunksize=1000):
    """
    Chunk generator. Take a CSV `reader` and yield
    `chunksize` sized slices.
    Source: https://gist.github.com/miku/820490
    """
    chunk = []
    for index, line in enumerate(reader):
        if index % chunksize == 0 and index > 0:
            yield chunk
            chunk = []
        chunk.append(line)
    yield chunk

def process(tsv_chunk, include_surrogate = False):
    local_counter = Counter()
    
    for row in tsv_chunk.copy():
        if include_surrogate:
            local_counter[(row[1], row[2], row[3])] += 1
        else:
            local_counter[row[1]] += 1
    
    return local_counter


@typechecked
def map_sample_protospacers(parsing_demult_handler, include_surrogate = False, cores=1):
    tsv_reader = csv.reader(parsing_demult_handler, delimiter='\t')  # change delimiter for normal csv files
    header = next(tsv_reader)
      
    read_chunks = gen_chunks(tsv_reader, chunksize=10000)

    process_func = partial(process, include_surrogate=include_surrogate)

    combined_counter = None
    if cores > 1:
        with mp.Pool(processes=cores) as pool:
            chunk_counters = pool.map(process_func, read_chunks)
            combined_counter = reduce(lambda x, y: x + y, chunk_counters)
    else:
        chunk_counters = [process_func(read_chunk) for read_chunk in read_chunks]
        combined_counter = reduce(lambda x, y: x + y, chunk_counters)

    return combined_counter 


###
### MAIN FUNCTIONS
###

'''
    Take in input FASTQ filename, and a set of whitelisted guide sequences
'''
@typechecked
def get_guide_counts_from_fastq(whitelist_guide_sequences_series: pd.Series, fastq_fn: str, hamming_threshold_strict: int = 3, hamming_threshold_dynamic: bool = False, parse_left_flank: bool = True, parse_flank_sequence: Union[None, str] = None, cores: int=1):
    # Retrieve all observed guide sequences
    print("Retrieving FASTQ guide sequences and counting: " + fastq_fn)
    observed_guide_raw_sequences = retrieve_fastq_guide_sequences(fastq_fn, parse_left_flank=parse_left_flank, parse_flank_sequence=parse_flank_sequence, cores=cores)
    
    # Count each unique observed guide sequence
    observed_guide_sequences_counts = Counter(observed_guide_raw_sequences)
    
    return get_guide_counts(observed_guide_sequences_counts, whitelist_guide_sequences_series, hamming_threshold_strict, hamming_threshold_dynamic, cores)


'''
    Take in input FASTQ filename, and a set of whitelisted guide sequences
'''
@typechecked
def get_guide_counts_from_reporter_tsv(whitelist_guide_sequences_series: pd.Series, reporter_tsv_fn: str, hamming_threshold_strict: int = 3, hamming_threshold_dynamic: bool = False, cores: int=1):
    combined_counter = None
    with open(reporter_tsv_fn, "r", newline='') as reporter_tsv_handler:
        combined_counter = map_sample_protospacers(reporter_tsv_handler, include_surrogate=False, cores=cores)

    return get_guide_counts(combined_counter, whitelist_guide_sequences_series, hamming_threshold_strict, hamming_threshold_dynamic, cores)

@typechecked
def get_reporter_counts_from_reporter_tsv(whitelist_guide_reporter_df: pd.DataFrame, reporter_tsv_fn: str, surrogate_hamming_threshold_strict: int = 10, barcode_hamming_threshold_strict: int = 2, hamming_threshold_strict: int = 7, hamming_threshold_dynamic: bool = False, cores: int=1):
    
    combined_counter = None
    with open(reporter_tsv_fn, "r", newline='') as reporter_tsv_handler:
        combined_counter = map_sample_protospacers(reporter_tsv_handler, include_surrogate = True, cores=cores)

    return get_reporter_counts(observed_guide_reporters_counts=combined_counter, whitelist_guide_reporter_df=whitelist_guide_reporter_df, surrogate_hamming_threshold_strict=surrogate_hamming_threshold_strict, barcode_hamming_threshold_strict=barcode_hamming_threshold_strict, hamming_threshold_strict=hamming_threshold_strict, hamming_threshold_dynamic=hamming_threshold_dynamic, cores=cores)


'''
    NOTE 8/28/223: Most of the code is located in Juypter notebook http://localhost:30111/notebooks/PROJECTS/2023_03_BB_SensorDemultiplexAndAnalysis/20230324_NovaSeq_Demultiplex.ipynb in anchor 20230828-millipede-encoding-anchor-new
'''
from typing import Tuple, Union, Mapping
from multiprocessing import Pool
from functools import partial
import pandas as pd
import numpy as np
import os
from functools import reduce
import pickle
from collections import defaultdict
from dataclasses import dataclass
from .CrisprAmbiguousMapping import save_or_load_pickle

def determine_mutations_in_sequence(true_sequence, observed_sequence):
    mutation_array = []
    lowest_len = min(len(true_sequence), len(observed_sequence))
    for i in range(lowest_len):
        if true_sequence[i] != observed_sequence[i]:
            preceding_nt_context = None
            succeeding_nt_context = None
            ref_nt = true_sequence[i]
            alt_nt = observed_sequence[i]
            if i == 0:
                preceding_nt_context = "_"
                succeeding_nt_context = true_sequence[i+1]
            elif i == len(true_sequence)-1:
                preceding_nt_context = true_sequence[i-1]
                succeeding_nt_context = "_"
            else:
                preceding_nt_context = true_sequence[i-1]
                succeeding_nt_context = true_sequence[i+1]
                
            mutation_series = pd.Series((preceding_nt_context, ref_nt, alt_nt,succeeding_nt_context, i), index=["preceding_nt_context", "ref_nt", "alt_nt", "succeeding_nt_context", "position"])
            mutation_array.append(mutation_series)
    observed_sequence_mutation_df = pd.DataFrame(mutation_array)
    return observed_sequence_mutation_df


# TODO 4/3/23: All this could probably be sped up through vectorization
def get_mutational_profile(observed_reporter_series, whitelist_reporter_tuple: Tuple[str, str, str]) -> Tuple[pd.DataFrame, pd.DataFrame]:
    observed_reporter_seq = observed_reporter_series["observed_sequence"]
    
    true_guide_seq = whitelist_reporter_tuple[0]
    true_surrogate_seq = whitelist_reporter_tuple[1]
    
    observed_guide_seq = observed_reporter_seq[0]
    observed_surrogate_seq = observed_reporter_seq[1]
    observed_surrogate_seq = observed_surrogate_seq[-len(true_surrogate_seq)-1:]
    
    observed_guide_mutation_df = determine_mutations_in_sequence(true_sequence=true_guide_seq, observed_sequence=observed_guide_seq)
    observed_surrogate_mutation_df = determine_mutations_in_sequence(true_sequence=true_surrogate_seq, observed_sequence=observed_surrogate_seq)

    return (observed_guide_mutation_df, observed_surrogate_mutation_df)


# Iterate through each whitelist reporter

'''
    Get the protospacer and surrogate mutation list for each whitelist entry. 
'''
def get_observed_mutation_per_whitelist_reporter(row, observed_guides_df):
    whitelist_reporter_tuple = tuple(row.values())
    
    observed_reporter_df = observed_guides_df[observed_guides_df["inferred_guides"]==whitelist_reporter_tuple]

    all_observed_guide_mutation_df_list = []
    all_observed_surrogate_mutation_df_list = []
    all_observed_reporter_series_mutation_count_series_list = []
    
    for _, observed_reporter_series in observed_reporter_df.iterrows():
        observed_count = observed_reporter_series["observed_counts"]
        observed_guide_mutation_df, observed_surrogate_mutation_df = get_mutational_profile(observed_reporter_series, whitelist_reporter_tuple)

        observed_guide_mutation_df["observed_count"] = observed_count
        observed_surrogate_mutation_df["observed_count"] = observed_count

        observed_guide_mutation_df["true_protospacer"] = row["protospacer"]
        observed_surrogate_mutation_df["true_protospacer"] = row["protospacer"]
        observed_surrogate_mutation_df["true_surrogate"] = row["surrogate"]

        # Get count of mutations per guide/surrogate
        
        observed_reporter_series_mutation_count_series = observed_reporter_series.copy()
        observed_reporter_series_mutation_count_series["guide_mutation_count"] = observed_guide_mutation_df.shape[0]
        observed_reporter_series_mutation_count_series["surrogate_mutation_count"] = observed_surrogate_mutation_df.shape[0]
        
        all_observed_guide_mutation_df_list.append(observed_guide_mutation_df)
        all_observed_surrogate_mutation_df_list.append(observed_surrogate_mutation_df)
        all_observed_reporter_series_mutation_count_series_list.append(observed_reporter_series_mutation_count_series)

    
    all_observed_guide_mutation_df = pd.concat(all_observed_guide_mutation_df_list) if len(all_observed_guide_mutation_df_list) != 0 else pd.DataFrame()
    all_observed_surrogate_mutation_df = pd.concat(all_observed_surrogate_mutation_df_list) if len(all_observed_surrogate_mutation_df_list) != 0 else pd.DataFrame()
    all_observed_reporter_series_mutation_count_df = pd.DataFrame(all_observed_reporter_series_mutation_count_series_list) if len(all_observed_reporter_series_mutation_count_series_list) != 0 else pd.DataFrame()

    return {
        "all_observed_guide_mutation_df": all_observed_guide_mutation_df,
        "all_observed_surrogate_mutation_df": all_observed_surrogate_mutation_df,
        "all_observed_reporter_series_mutation_count_df": all_observed_reporter_series_mutation_count_df
    }


# define a function to apply multiprocessing
'''
    Given the count result, generate the dataframes summarizing the mutations in the observed sequences
'''
def get_observed_mutation_dfs(count_result: Tuple[pd.Series, pd.DataFrame, Mapping], cores=1):
    whitelist_guide_reporter_counts, observed_guides_df, qc_dict = count_result
    whitelist_guide_reporter_df = whitelist_guide_reporter_counts.index.to_frame().reset_index(drop=True)

    # define the number of processes to use

    # create a Pool object with the number of processes defined above
    
    results = None
    
    get_observed_mutation_per_whitelist_reporter_p = partial(get_observed_mutation_per_whitelist_reporter, observed_guides_df=observed_guides_df)
    if cores == 1:
        results = [get_observed_mutation_per_whitelist_reporter_p(record) for record in whitelist_guide_reporter_df.to_dict("records")]
    else:
        print(f"Running parallelized with {cores} cores")
        with Pool(processes=cores) as pool:
            # apply the process_row function to each row of the whitelist_guide_reporter_df DataFrame using map()
            results = pool.map(get_observed_mutation_per_whitelist_reporter_p, whitelist_guide_reporter_df.to_dict("records"))
    
    # concatenate all the resulting DataFrames
    all_observed_guide_mutation_df = pd.concat([r["all_observed_guide_mutation_df"] for r in results])
    all_observed_surrogate_mutation_df = pd.concat([r["all_observed_surrogate_mutation_df"] for r in results])
    all_observed_reporter_series_mutation_count_df = pd.concat([r["all_observed_reporter_series_mutation_count_df"] for r in results], axis=0)

    return {
        "all_observed_guide_mutation_df": all_observed_guide_mutation_df,
        "all_observed_surrogate_mutation_df": all_observed_surrogate_mutation_df,
        "all_observed_reporter_series_mutation_count_df": all_observed_reporter_series_mutation_count_df
    }


'''
function copied from "anchor 20230703-millipede-encoding-anchor" in notebook "2021_11_BB_Shared_Tiling_Screen_Analysis/20220504_Davide_pilot_BE_analysis/STEP%202-%20Prepare%20base-editing%20allele%20encoding%20.ipynb"

TODO: Once we build the package for crispr-millipede, replace with using the function from crispr-millipede. There are a few differences, specifically had to change np.object and np.str to np.object_ and np.str_ probably due to numpy version differences. Also added N and changed assertion statement

'''
def get_substitution_encoding(aligned_sequence, original_seq, skip_index=0):
    assert len(aligned_sequence) == len(original_seq) # Ensure the aligned sequence (from allele table) is equal size to the reference sequence
    
    
    nucleotides = ["A","C","T","G","N","-"] # List of possible nucleotides
    encodings_per_position = []
    mismatch_mappings_per_position = []
    for index in range(0, len(original_seq)): # Iterate through each base and check for substitution
        # TODO Ensure sequences are uppercase
        nucleotides_mm = nucleotides[:]
        nucleotides_mm.remove(original_seq[index])
        mm_encoding = pd.Series(np.repeat(0, len(nucleotides)-1))
        if aligned_sequence[index] == original_seq[index]: # If the aligned sequence is same as reference
            pass
        else:
            mm_index = nucleotides_mm.index(aligned_sequence[index])
            mm_encoding[mm_index] = 1
        mismatch_mappings_per_position.append(nucleotides_mm)
        encodings_per_position.append(mm_encoding)

    encodings_per_position_df = pd.DataFrame(encodings_per_position).T
    mismatch_mappings_per_position_df = pd.DataFrame(mismatch_mappings_per_position).T

    encodings_per_position_df.columns = list(original_seq)
    mismatch_mappings_per_position_df.columns = list(original_seq)

    mismatch_mappings_per_position_POS_list = np.arange(mismatch_mappings_per_position_df.shape[1]).repeat(mismatch_mappings_per_position_df.shape[0])
    mismatch_mappings_per_position_REF_list = np.asarray(list(original_seq)).repeat(mismatch_mappings_per_position_df.shape[0]).astype(np.object_)
    mismatch_mappings_per_position_ALT_list = mismatch_mappings_per_position_df.T.values.flatten()
    mismatch_mappings_per_position_full_list = mismatch_mappings_per_position_POS_list.astype(np.str_).astype(object)+mismatch_mappings_per_position_REF_list + np.repeat(">", len(mismatch_mappings_per_position_REF_list)) + mismatch_mappings_per_position_ALT_list
    encodings_per_position_list = encodings_per_position_df.T.values.flatten()

    
    # Encodings per position DF, mismatch mappings per position DF, encodings per position flattened, mismatch mappings per position flattened, mismatch mapping position in flattened list, mismatch mapping ref in flattened list, mismatch mapping alt in flattened list, all substitutions made
    
    index = pd.MultiIndex.from_tuples(zip(mismatch_mappings_per_position_full_list, mismatch_mappings_per_position_POS_list, mismatch_mappings_per_position_REF_list, mismatch_mappings_per_position_ALT_list), names=["FullChange", "Position","Ref", "Alt"])
    
    assert len(encodings_per_position_list) == len(index), f"{len(encodings_per_position_list)} == {len(index)}"
    encodings_per_position_series = pd.Series(encodings_per_position_list, index = index, name="encoding")
    return encodings_per_position_series


def process_encoding(encoding_set):
    for encoding_df in encoding_set:
        encoding_df.columns = encoding_df.columns.get_level_values("FullChange")
        
        
def add_read_column(original_dfs, encoded_dfs, suffix):
    for i, original_dfs_rep in enumerate(original_dfs):
        encoded_dfs[i]["#Reads{}".format(suffix)] = original_dfs_rep["#Reads"]
        
        
def collapse_encodings(encoded_dfs):
    encoded_dfs_collapsed = []
    for encoded_df_rep in encoded_dfs:
        feature_colnames = [name for name in list(encoded_df_rep.columns) if "#Reads" not in name]
        encoded_dfs_collapsed.append(encoded_df_rep.groupby(feature_colnames, as_index=True).sum().reset_index())
    return encoded_dfs_collapsed


def merge_conditions_by_rep(first_encodings_collapsed, second_encodings_collapsed, third_encodings_collapsed):
    assert len(first_encodings_collapsed) == len(second_encodings_collapsed) == len(third_encodings_collapsed)
    encoded_dfs_merged = []
    for rep_i in range(len(first_encodings_collapsed)):
        feature_colnames |= [name for name in list(first_encodings_collapsed[rep_i].columns) if "#Reads" not in name]
        samples = [first_encodings_collapsed[rep_i], second_encodings_collapsed[rep_i], third_encodings_collapsed[rep_i]]
        df_encoding_rep1 = reduce(lambda  left,right: pd.merge(left,right,on=feature_colnames,
                                                how='outer'), samples).fillna(0)
        
        encoded_dfs_merged.append(df_encoding_rep1)
    return encoded_dfs_merged


@dataclass
class GuideReporterEncodingStrategy:
    output_dir: str
    date_string: str


def merge_and_encode_observed_surrogate(args):
    true_surrogate_sequence, condition_dataframes = args
    dataframes = [dataframe for condition, dataframe in condition_dataframes]
    
    if len(dataframes) > 1:
        merged_observed_surrogate_sequence_df = reduce(lambda left, right: pd.merge(left, right, how="outer", on=["observed_protospacer", "observed_surrogate"]), dataframes)
    else:
        merged_observed_surrogate_sequence_df = pd.DataFrame(dataframes[0])
    
    merged_observed_surrogate_sequence_df = merged_observed_surrogate_sequence_df.fillna(0.0)

    merged_observed_surrogate_sequence_df["observed_surrogate"] = merged_observed_surrogate_sequence_df["observed_surrogate"].apply(lambda sequence: sequence[-len(true_surrogate_sequence[1]):])
    merged_observed_surrogate_sequence_df = merged_observed_surrogate_sequence_df.groupby(["observed_protospacer", "observed_surrogate"]).sum()
    merged_observed_surrogate_sequence_df["edited_protospacer"] = merged_observed_surrogate_sequence_df.index.get_level_values('observed_protospacer') != true_surrogate_sequence[0]
    merged_observed_surrogate_sequence_df["edited_surrogate"] = merged_observed_surrogate_sequence_df.index.get_level_values('observed_surrogate') != true_surrogate_sequence[1]
    merged_observed_surrogate_sequence_df = merged_observed_surrogate_sequence_df.sort_values(f"observed_counts_{condition_dataframes[0][0]}", ascending=False)
    
    merged_observed_surrogate_sequence_surrogate_encoding_df = None
    if len(merged_observed_surrogate_sequence_df.index) > 0:
        merged_observed_surrogate_sequence_surrogate_encoding_df = pd.concat([get_substitution_encoding(observed_surrogate, true_surrogate_sequence[1]) for observed_surrogate in merged_observed_surrogate_sequence_df.index.get_level_values('observed_surrogate')], axis=1).transpose()
        nt_columns = merged_observed_surrogate_sequence_surrogate_encoding_df.columns
        merged_observed_surrogate_sequence_surrogate_encoding_df.index = merged_observed_surrogate_sequence_df.index.get_level_values('observed_surrogate')

        for colname in merged_observed_surrogate_sequence_df.columns:
            merged_observed_surrogate_sequence_surrogate_encoding_df[colname] = merged_observed_surrogate_sequence_df[colname].values
        
        all_columns = merged_observed_surrogate_sequence_surrogate_encoding_df.columns
        # Can only groupby on single index, so temporarily setting column as FullChange index
        merged_observed_surrogate_sequence_surrogate_encoding_df.columns = list(merged_observed_surrogate_sequence_surrogate_encoding_df.columns.get_level_values("FullChange"))
        merged_observed_surrogate_sequence_surrogate_encoding_df_groups = merged_observed_surrogate_sequence_surrogate_encoding_df.groupby(list(nt_columns.get_level_values("FullChange")), as_index=False)
        group_index = [group.index[0] for _, group in merged_observed_surrogate_sequence_surrogate_encoding_df_groups]
        # TODO: The edited_surrogate and edited_protospacer are edited as well
        merged_observed_surrogate_sequence_surrogate_encoding_df = pd.concat([group.sum() for _, group in merged_observed_surrogate_sequence_surrogate_encoding_df_groups], axis=1).transpose()
        merged_observed_surrogate_sequence_surrogate_encoding_df.index = group_index
        merged_observed_surrogate_sequence_surrogate_encoding_df.columns = all_columns

    merged_observed_surrogate_sequence_protospacer_encoding_df = None
    if len(merged_observed_surrogate_sequence_df.index) > 0:
        merged_observed_surrogate_sequence_protospacer_encoding_df = pd.concat([get_substitution_encoding(observed_protospacer, true_surrogate_sequence[0]) for observed_protospacer in merged_observed_surrogate_sequence_df.index.get_level_values('observed_protospacer')], axis=1).transpose()
        nt_columns = merged_observed_surrogate_sequence_protospacer_encoding_df.columns
        merged_observed_surrogate_sequence_protospacer_encoding_df.index = merged_observed_surrogate_sequence_df.index.get_level_values('observed_protospacer')

        for colname in merged_observed_surrogate_sequence_df.columns:
            merged_observed_surrogate_sequence_protospacer_encoding_df[colname] = merged_observed_surrogate_sequence_df[colname].values
        
        all_columns = merged_observed_surrogate_sequence_protospacer_encoding_df.columns
        # Can only groupby on single index, so temporarily setting column as FullChange index
        merged_observed_surrogate_sequence_protospacer_encoding_df.columns = list(merged_observed_surrogate_sequence_protospacer_encoding_df.columns.get_level_values("FullChange"))
        merged_observed_surrogate_sequence_protospacer_encoding_df_groups = merged_observed_surrogate_sequence_protospacer_encoding_df.groupby(list(nt_columns.get_level_values("FullChange")), as_index=False)
        group_index = [group.index[0] for _, group in merged_observed_surrogate_sequence_protospacer_encoding_df_groups]
        merged_observed_surrogate_sequence_protospacer_encoding_df = pd.concat([group.sum() for _, group in merged_observed_surrogate_sequence_protospacer_encoding_df_groups], axis=1).transpose()
        merged_observed_surrogate_sequence_protospacer_encoding_df.index = group_index
        merged_observed_surrogate_sequence_protospacer_encoding_df.columns = all_columns

        
        
    return true_surrogate_sequence, merged_observed_surrogate_sequence_df, merged_observed_surrogate_sequence_surrogate_encoding_df, merged_observed_surrogate_sequence_protospacer_encoding_df

    
def process_tiling_reporter_counts(sample_pooling_guidelibrary_df: pd.DataFrame, date_string: str, screen_column_group: str, replicate_column: Union[int, str], condition_column, guide_reporter_encoding_strategy, cores=1):
    # Modify below:
    surrogate_output_sub_dir = f"{guide_reporter_encoding_strategy.output_dir}"

    if not os.path.exists(surrogate_output_sub_dir):
        os.makedirs(surrogate_output_sub_dir)
        
    merged_observed_surrogate_sequence_df_dict_screen_reps_library = {}
    merged_observed_surrogate_sequence_surrogate_encoding_df_dict_screen_reps_library = {}
    merged_observed_surrogate_sequence_protospacer_encoding_df_dict_screen_reps_library = {}
    for guide_library_id, library_group in sample_pooling_guidelibrary_df.groupby("guide_library_id"):
        print(f"Processing guide library {guide_library_id}")
        
        merged_observed_surrogate_sequence_df_dict_screen_reps = {}
        merged_observed_surrogate_sequence_surrogate_encoding_df_dict_screen_reps = {}
        merged_observed_surrogate_sequence_protospacer_encoding_df_dict_screen_reps = {}
            
        # Iterate through each screen group
        for screen_name, screen_group in library_group.groupby(screen_column_group):
            print(f"Processing screen {screen_name}")

            merged_observed_surrogate_sequence_df_dict_screen_rep = {}
            merged_observed_surrogate_sequence_surrogate_encoding_df_dict_screen_rep = {}
            merged_observed_surrogate_sequence_protospacer_encoding_df_dict_screen_rep = {}
            for replicate, replicate_group in screen_group.groupby(replicate_column):
                print(f"Processing replicate {replicate}")

                condition_observed_surrogate_sequence_df_dict = defaultdict(list)

                # Iterate through each screen condition
                conditions = list(set(replicate_group[condition_column].astype(str)))
                for condition in conditions:
                    print(f"Processing condition {condition}")
                    # Get the barcode for the screen condition sample
                    
                    # Get the metadata needed to construct the filename of the count matrix file
                    i5_index = replicate_group.loc[replicate_group[condition_column].astype(str) == condition, "i5_index"].iloc[0]
                    sample_sub_dir = sample_indices_df_subset.loc[sample_indices_df_subset["i5_index"] == i5_index,"sub_library_name"].iloc[0]
                    U6PE1_barcode = replicate_group.loc[replicate_group[condition_column].astype(str) == condition, "U6PE1_Barcode"].iloc[0]
                    with open(f"{guide_reporter_counting_strategy.output_dir}/{sample_sub_dir}/{sample_sub_dir}_#{U6PE1_barcode}_count_{date_string}.pickle", 'rb') as handle:

                        count_result = pickle.load(handle)

                        whitelist_guide_reporter_counts, observed_guides_df, qc_dict = count_result
                        whitelist_guide_reporter_list = list(whitelist_guide_reporter_counts.index)

                        # Get the surrogate count df for the screen condition sample
                        for true_surrogate_sequence in whitelist_guide_reporter_list:
                            observed_surrogate_info_sequence_df = observed_guides_df[observed_guides_df["inferred_guides"] == true_surrogate_sequence].loc[:,["observed_sequence", "observed_counts"]]
                            observed_surrogate_sequence_df = pd.DataFrame({"observed_protospacer": observed_surrogate_info_sequence_df["observed_sequence"].apply(lambda info: info[0]), "observed_surrogate": observed_surrogate_info_sequence_df["observed_sequence"].apply(lambda info: info[1]), f"observed_counts_{condition}": observed_surrogate_info_sequence_df["observed_counts"]})
                            condition_observed_surrogate_sequence_df_dict[true_surrogate_sequence].append((condition, observed_surrogate_sequence_df))

                # TODO: For cores == 1, using regular list comprehension. Not sure how to deal with the two dicts that need updating, perhaps modify function to return if cores==1
                # Create the encoding for each whitelist surrogate (parallelized)
                print(f"Performing encoding of {replicate} with cores {cores}")
                if cores > 1:
                    with Pool(processes=cores) as pool:
                        # Prepare the arguments for the parallel processing
                        args_list = [(true_surrogate_sequence, condition_dataframes)
                                     for true_surrogate_sequence, condition_dataframes in condition_observed_surrogate_sequence_df_dict.items()]

                        # Use the Pool's map function to process the loop iterations in parallel
                        results = pool.map(merge_and_encode_observed_surrogate, args_list)
                else:
                    # Prepare the arguments for the parallel processing
                    args_list = [(true_surrogate_sequence, condition_dataframes)
                                 for true_surrogate_sequence, condition_dataframes in condition_observed_surrogate_sequence_df_dict.items()]

                    # Use the Pool's map function to process the loop iterations in parallel
                    results = [merge_and_encode_observed_surrogate(args) for args in args_list]
                print(f"Completed encoding of {replicate}. Processing results") 
                # Parse the parallelized result into the dictionaries
                merged_observed_surrogate_sequence_df_dict = {} # Contains counts per condition samples
                merged_observed_surrogate_sequence_surrogate_encoding_df_dict = {} # Contain encodings per condition samples
                merged_observed_surrogate_sequence_protospacer_encoding_df_dict = {} # Contain encodings per condition samples
                for true_surrogate_sequence, merged_observed_surrogate_sequence_df, merged_observed_surrogate_sequence_surrogate_encoding_df, merged_observed_surrogate_sequence_protospacer_encoding_df in results:
                    merged_observed_surrogate_sequence_df_dict[true_surrogate_sequence] = merged_observed_surrogate_sequence_df
                    merged_observed_surrogate_sequence_surrogate_encoding_df_dict[true_surrogate_sequence] = merged_observed_surrogate_sequence_surrogate_encoding_df
                    merged_observed_surrogate_sequence_protospacer_encoding_df_dict[true_surrogate_sequence] = merged_observed_surrogate_sequence_protospacer_encoding_df

                # Add dictionaries to main list
                merged_observed_surrogate_sequence_df_dict_screen_rep[replicate] = merged_observed_surrogate_sequence_df_dict
                merged_observed_surrogate_sequence_surrogate_encoding_df_dict_screen_rep[replicate] = merged_observed_surrogate_sequence_surrogate_encoding_df_dict
                merged_observed_surrogate_sequence_protospacer_encoding_df_dict_screen_rep[replicate] = merged_observed_surrogate_sequence_protospacer_encoding_df_dict

            merged_observed_surrogate_sequence_df_dict_screen_reps[screen_name] = merged_observed_surrogate_sequence_df_dict_screen_rep
            merged_observed_surrogate_sequence_surrogate_encoding_df_dict_screen_reps[screen_name] = merged_observed_surrogate_sequence_surrogate_encoding_df_dict_screen_rep
            merged_observed_surrogate_sequence_protospacer_encoding_df_dict_screen_reps[screen_name] = merged_observed_surrogate_sequence_protospacer_encoding_df_dict_screen_rep
        
        print(f"Completed guide library encoding: {guide_library_id}. Now saving.")
        save_or_load_pickle(f"{surrogate_output_sub_dir}/", f"merged_observed_surrogate_sequence_df_dict_screen_reps_{guide_library_id}", py_object=merged_observed_surrogate_sequence_df_dict_screen_reps, date_string=guide_reporter_encoding_strategy.date_string)
        save_or_load_pickle(f"{surrogate_output_sub_dir}/", f"merged_observed_surrogate_sequence_surrogate_encoding_df_dict_screen_reps_{guide_library_id}", py_object=merged_observed_surrogate_sequence_surrogate_encoding_df_dict_screen_reps, date_string=guide_reporter_encoding_strategy.date_string)
        save_or_load_pickle(f"{surrogate_output_sub_dir}/", f"merged_observed_surrogate_sequence_protospacer_encoding_df_dict_screen_reps_{guide_library_id}", py_object=merged_observed_surrogate_sequence_protospacer_encoding_df_dict_screen_reps, date_string=guide_reporter_encoding_strategy.date_string)
    
        merged_observed_surrogate_sequence_df_dict_screen_reps_library[guide_library_id] = merged_observed_surrogate_sequence_df_dict_screen_reps
        merged_observed_surrogate_sequence_surrogate_encoding_df_dict_screen_reps_library[guide_library_id] = merged_observed_surrogate_sequence_surrogate_encoding_df_dict_screen_reps
        merged_observed_surrogate_sequence_protospacer_encoding_df_dict_screen_reps_library[guide_library_id] = merged_observed_surrogate_sequence_protospacer_encoding_df_dict_screen_reps

    print("Completed all encoding, saving results")
    save_or_load_pickle(f"{surrogate_output_sub_dir}/", "merged_observed_surrogate_sequence_df_dict_screen_reps_library", py_object=merged_observed_surrogate_sequence_df_dict_screen_reps_library, date_string=guide_reporter_encoding_strategy.date_string)
    save_or_load_pickle(f"{surrogate_output_sub_dir}/", "merged_observed_surrogate_sequence_surrogate_encoding_df_dict_screen_reps_library", py_object=merged_observed_surrogate_sequence_surrogate_encoding_df_dict_screen_reps_library, date_string=guide_reporter_encoding_strategy.date_string)
    save_or_load_pickle(f"{surrogate_output_sub_dir}/", "merged_observed_surrogate_sequence_protospacer_encoding_df_dict_screen_reps_library", py_object=merged_observed_surrogate_sequence_protospacer_encoding_df_dict_screen_reps_library, date_string=guide_reporter_encoding_strategy.date_string)

    return merged_observed_surrogate_sequence_df_dict_screen_reps_library, merged_observed_surrogate_sequence_surrogate_encoding_df_dict_screen_reps_library, merged_observed_surrogate_sequence_protospacer_encoding_df_dict_screen_reps_library
